%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% A routine to compare l1, SBL and OMP for sinusoidal signals recovery
% sampled using random linear projections
%
% Implements the example in: S. Engelberg, "Compressive Sensing", IEEE
% Instrumentation and Measurement Magazine, Feb. 2012.
%
% Dependencies: l1 magic package, OMP package
%
% Author: Chandra Murthy, ECE Dept., IISc
% Email : cmurthy@iisc.ac.in
% Date last updated : 24 Jan. 2020
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


clear all; close all;

% add relevant paths
% Replace this with the path to the location of CVX
%addpath('../../cvx')
addpath('./l1magic', './l1magic/optimization')

% number of measurements
m = 40;
% number of samples in 1 second, also, length of sparse signal
n = 512;

% List of tones that go into the sinusoidal signal
k = 1; % number of tones; it is a measure of sparsity
f = [10]; 100*rand([1,k]); % Hz
theta = 2*pi*rand([1,k]); % radians
amp = abs(10*randn([1,k])); % value
%amp = 10*ones([1,k]);

% sparsity: length(f) is only the approximate value of sparsity since we
% are going to use windowing
k = length(f);

% time domain samples
y = sum((ones(n,1)*amp).*sin(2*pi*[0:1:n-1]'*f/n + ones(n,1)*theta),2);

% FFT and inverse FFT matrices
F = fft(eye(n));
iF = F'/n;

% apply a Hanning window to smooth out the edges
W = diag(hanning(n));
ywin = W*y;

plot([0:n-1]/n, y, 'r-', [0:n-1]/n, ywin, 'b-');
xlabel('Time');
ylabel('Time-domain Signal');
legend('Original signal', 'windowed signal');

%pause;
%close;

% frequency domain vector, will be more or less sparse
Y = F'*ywin/n;

plot(([0:1:n-1]-n/2)/2, abs([Y(n/2+1:n); Y(1:n/2)]), 'r*');
legend('Orig. windowed sinusoid');
xlabel('Frequency (Hz)');
ylabel('Frequency response');
ylim([-1,max(amp)+1])

%pause;
%close;

% measurement matrix phi, observation b
phi = randn(m,n);
b = phi*ywin;

% Overall measurement matrix; b = phi*F*F'*ywin, and F'*ywin is sparse.
A = phi*F;

x0 = A'*b/(m*n);
bigA = [real(A), -imag(A); imag(A), real(A)];
bigx0 = [real(x0); imag(x0)];
bigb = [real(b); imag(b)];


x0 = bigA'*bigb;

% solve the LP
%tic
bigYpcvx = l1eq_pd(x0, bigA, [], bigb, 10^(1));
%toc
Ypcvx = bigYpcvx(1:n) + 1i*bigYpcvx(n+1:2*n);
ypcvx = real(F*Ypcvx);


%% Solve l2 recovery

% Least-squares solution from b
Ypmmse = (A'*inv(A*A'))*b;
ypmmse = real(F*Ypmmse);

% sub-sample and interpolate y
y_sub = y(1:floor(n/m):end);
length_y_sub = length(y_sub)
y_interp = interp(y_sub,floor(n/m));
y_interp_win = diag(hanning(length(y_interp)))*y_interp;


%% Solve OMP
s = min(10*k,m);
Ypomp = OMP(A, b, s, []);
ypomp = real(F*Ypomp);



figure;
% need to take care of the wrap-around
plot(([0:1:n-1]-n/2)/2, abs([Y(n/2+1:n); Y(1:n/2)]), 'r*',...
    ([0:1:n-1]-n/2)/2, abs([Ypcvx(n/2+1:n); Ypcvx(1:n/2)]), 'gd',...
    ([0:1:n-1]-n/2)/2, abs([Ypmmse(n/2+1:n); Ypmmse(1:n/2)]), 'mx',...
    ([0:1:n-1]-n/2)/2, abs([Ypomp(n/2+1:n); Ypomp(1:n/2)]), 'kp');
legend('Orig. windowed sinusoid', ...
    'L1-min','L2-min','OMP');
xlabel('Frequency (Hz)');
ylabel('y');
title(['num. obsvns = ', num2str(m), ...
    ', num. vars = ', num2str(n)]);
ylim([-1,max(amp)+1])


figure;
plot([0:1:n-1]/n, ywin(1:1:n), 'r-',...
    [0:1:n-1]/n, ypcvx(1:1:n), 'g.-',...
    [0:1:n-1]/n, ypmmse(1:1:n), 'm--',...
    [0:1:length(y_interp_win)-1]/length(y_interp_win), y_interp_win, 'b:',...
    [0:1:n-1]/n, ypomp(1:1:n), 'k.');
legend('Orig. windowed sinusoid', ...
    'L1-min','L2-min','resample','OMP');
xlabel('time');
ylabel('y');
title(['num. obsvns = ', num2str(m), ...
    ', num. vars = ', num2str(n)]);
ylim([-max(amp)-1,max(amp)+1]);

disp('Frequency domain:');
disp(['MSE(L1) = ', num2str(sum(abs(Y-Ypcvx).^2)/n,2), ...
    ', MSE(L2) = ', num2str(sum(abs(Y-Ypmmse).^2)/n,2), ...
    ', MSE(OMP) = ', num2str(sum(abs(Y-Ypomp).^2)/n,2)]);

disp('Time domain:');
disp(['MSE(L1) = ', num2str(sum(abs(ywin-ypcvx).^2)/n,2), ...
    ', MSE(L2) = ', num2str(sum(abs(ywin-ypmmse).^2)/n,2), ...
    ', MSE(OMP) = ', num2str(sum(abs(ywin-ypomp).^2)/n,2)]);

